#########################################################################
# qq.tcl                                                                #
# graph: simple interactive graph editor to design the relations among  #
#        subpopulations                                                 #
#########################################################################
set w .relation
catch {destroy $w}
toplevel $w
wm title $w "Relation of subpopulations"
wm iconname $w "relation"
positionWindow3 $w

canvas $w.c -width 12.5c -height 11c
pack $w.c

global env

global ccnode
global sorted_list
global dd 
global ll 
global dd_num
set n 0
set dd_num 0
set dd {}
set dd1 {}
set ll {}
set gn {}
set node_X {}
set node_Y {}
set sorted_list {}

frame $w.buttons
button $w.buttons.dismiss -text "Dismiss" -command "destroy $w"
pack  $w.buttons -side bottom -expand y -fill x -pady 2m
button $w.buttons.unset -text "Clear" -command {
 source $env(GALOPPS_GUI_HOME)\/relation.tcl
}
button $w.buttons.code -text "Set Migration" -command {
 get_list
 source $env(GALOPPS_GUI_HOME)\/migration.tcl
 mkInitial
 mkRelation $sorted_list 
}

#
# NEW addition of Help button on 7/15/96 by Owen Mathews and Erik Goodman
#

button $w.buttons.help -text "Help" -command "helpDialog nodes.help"
pack $w.buttons.help $w.buttons.dismiss $w.buttons.unset $w.buttons.code -side left -expand 1

#
# END new addition
#

############################################################
# get the list with all relation between subpopulation     #
# from here call mksortedlist that make the raw list sorted#
############################################################
proc get_list {} {
 global dd 
 global dd_num 
 global dd1
 global ll
 global new_list
 global sorted_list
 global nodeX nodeY edgeFirst edgeSecond

 set nk {}
 set nk_new {}

 foreach l $ll {
   foreach d $dd {
      set k1 [lsearch $edgeFirst($d) $l]
      if { $k1 != -1} {
       set nk [concat $nk $d]
       set ls $l
       }
   }
   foreach d $dd {
      set k2 [lsearch $edgeSecond($d) $ls]
      if {$k2 != -1} {
       set nk [concat $nk $d]
       }
      }
    }

   foreach i $nk {
    set zh [lsearch $dd $i]
    set kk [lindex $dd1 $zh]
    set nk_new [concat $nk_new $kk]
    }
   #puts stdout $nk_new
   set sorted_list [mksortedlist $nk_new]
#   puts stdout $sorted_list
}

###############################################################
# make the raw list (all_subpopulation ) sorted that meen from# 
# sub0 to sub(papm(app_npops))                                #
# from here call mksortedlist that make the raw list sorted   #
###############################################################
proc mksortedlist list {
 global new_list
 set new_list {}
 set length_list [llength $list]
 set l [lrange $list 0 1]
 set list [lreplace $list 0 1]
 set new_length [llength $list]
 set time1 $length_list
 set time2 [expr $new_length/2]

 for {set j 0} {$j < $time1} {incr j 2} { 
  if {$length_list == 2} {
   set new_list [concat $new_list $l]
   return $new_list
   break
   }
 
 for {set i 0} {$i < $new_length} {incr i 2} {
   set first_l [lindex $l 0]
   set second_l [lindex $l 1] 
   set first_list [lindex $list $i]
   set next [expr $i + 1]
   set second_list [lindex $list $next]
  if {$first_list < $first_l} {
     set ls $l
     set l [lrange $list $i $next]
     set first_ls [lindex $ls 0]
     set second_ls [lindex $ls 1]
     set list [lreplace $list $i $i $first_ls] 
     set list [lreplace $list $next $next $second_ls] 
     }
   if {($first_list == $first_l) && ($second_list < $second_l)} {
     set ls $l
     set l [lrange $list $i $next]
     set first_ls [lindex $ls 0]
     set second_ls [lindex $ls 1]
     set list [lreplace $list $i $i $first_ls]
     set list [lreplace $list $next $next $second_ls]
     }
   }
 set new_list [concat $new_list $l]
 set length_list [llength $list]
 set l [lrange $list 0 1]
 set list [lreplace $list 0 1]
 set new_length [llength $list]
 }
}

###################################################################
# draw node on screen and register it in a list also draw number  #
# attached the node meenwhile register the number in another list #
###################################################################
proc mkNode {x y} { 

 global dd 
 global dd1 
 global dd_num
 global n_last
 global n gn nd 
 global gnew
 global new
 global g_last_num g_last_node
 global nodeX nodeY edgeFirst edgeSecond
 global node_X node_Y 

 set new [.relation.c create oval [expr $x-10] [expr $y-10] \
         [expr $x+10] [expr $y+10] -outline black \
         -fill white -tags node]
 set nodeX($new) $x
 set nodeY($new) $y
 set edgeFirst($new) {}
 set edgeSecond($new) {}

 set gnew [.relation.c create text [expr $x] [expr $y] -text "s${n}" -tags nnode \
	  -anchor center] 
 set nd($gnew) $new
 set gn [concat $gn $gnew]

 set dd [concat $dd $new]
 set dd1 [concat $dd1 $dd_num]

 set node_X [concat $node_X $x]
 set node_Y [concat $node_Y $y]

 incr dd_num

 set n_last $n
 incr n
}

###############################################################
# Create edge between nods and register every edge in a list  #
###############################################################
proc mkEdge {first second} {        

 global ll
 global nodeX nodeY edgeFirst edgeSecond
 #puts stdout $first
 #puts stdout $second

 set edge [.relation.c create line $nodeX($first) $nodeY($first) \
	   $nodeX($second) $nodeY($second)  -arrow first -arrowshape {18 24 6} \
	   -tags gline]

 .relation.c lower $edge
 lappend edgeFirst($first) $edge
 lappend edgeSecond($second) $edge

 set ll [concat $ll $edge]
}

###############################################################
# move the number in dode                                     #
###############################################################
proc mmoveNode {node xDist yDist} {

  global nodeX nodeY edgeFirst edgeSecond
  .relation.c move $node $xDist $yDist
}

###############################################################
# move the node on screen                                     #
###############################################################
proc moveNode {node xDist yDist} {

  global nodeX nodeY edgeFirst edgeSecond
  .relation.c move $node $xDist $yDist
  incr nodeX($node) $xDist
  incr nodeY($node) $yDist
  foreach edge $edgeFirst($node) {
    .relation.c coords $edge $nodeX($node) $nodeY($node) \
         [lindex [.relation.c coords $edge] 2]\
         [lindex [.relation.c coords $edge] 3]
  }
  foreach edge $edgeSecond($node) {
    .relation.c coords $edge [lindex [.relation.c coords $edge] 0]\
	 [lindex [.relation.c coords $edge] 1]\
         $nodeX($node) $nodeY($node)
  }
}

#############################
# to deal every bind event  #
#############################
bind .relation.c <Button-1> {mkNode %x %y}

.relation.c bind nnode <Any-Enter> {
  .relation.c itemconfigure current -fill red
}

.relation.c bind nnode <Any-Leave> {
   .relation.c itemconfigure current -fill black
}

.relation.c bind gline <Any-Enter> {
   .relation.c itemconfigure current -fill white
}

.relation.c bind gline <Any-Leave> {
   .relation.c itemconfigure current -fill black
}

bind .relation.c 1 {
 set c  [.relation.c find withtag current]
 set c1 [lsearch $dd $c]
 set c2 [lsearch $gn $c]
  if { $c1 != -1} {
   } else {
       set firstNode [lindex $dd $c2]
       }
}

bind .relation.c 2 {
 set c  [.relation.c find withtag current]
 set c1 [lsearch $dd $c]
 set c2 [lsearch $gn $c]
   if { $c1 != -1} {
     set curNode $c
     } else {
        set curNode [lindex $dd $c2]
        }
 set signal 1

 if {($firstNode !="") && ($curNode !="")} {
 if { $ll == {} } { mkEdge $firstNode $curNode }

 if { $ll != {} } {
  foreach exit_line $ll {
   set in_edgeFirst [lsearch $edgeFirst($firstNode) $exit_line]
   set in_edgeSecond [lsearch $edgeSecond($curNode) $exit_line]
   if { ($in_edgeFirst != -1) && ($in_edgeSecond != -1)} {
     set signal -1
   }
  }
 }

 if { $signal == 1 } { mkEdge $firstNode $curNode }
 }
}

.relation.c bind nnode <Button-2> {
 set curX %x
 set curY %y
}

.relation.c bind node <B2-Motion> {
  set curX %x
  set curY %y
}

.relation.c bind nnode <B2-Motion> {
 global ccnode
 set ccnode  [.relation.c find withtag current]
 
 set num_index [lsearch $gn $ccnode]
# puts stdout $num_index
# puts stdout %x
 
 set node_X [lreplace $node_X $num_index $num_index %x] 
 set node_Y [lreplace $node_Y $num_index $num_index %y]

 moveNode $nd($ccnode) [expr %x-$curX] [expr %y-$curY]
 mmoveNode $ccnode [expr %x-$curX] [expr %y-$curY]

 set curX %x
 set curY %y
}

.relation.c bind gline  <Button-3> {
 set ggline [.relation.c find withtag current]
 set l [.relation.c find withtag current]

 .relation.c delete $ggline 

 set l1 [lsearch $ll $l]
 set ll [lreplace $ll $l1 $l1]

 foreach d $dd {
  set l1 [lsearch $edgeFirst($d) $l]
  if { $l1 != -1 } {
    set edgeFirst($d) [lreplace $edgeFirst($d) $l1 $l1]
  }
 }
 foreach d $dd {
  set l1 [lsearch $edgeSecond($d) $l]
  if { $l1 != -1 } {
    set edgeSecond($d) [lreplace $edgeSecond($d) $l1 $l1]
  }
 }
}

.relation.c bind nnode  <Button-3> {
 set d_line1 {}
 set d_line2 {}
 set d_line3 {}

 set c  [.relation.c find withtag current]
 set c1 [lsearch $dd $c]
 set c2 [lsearch $gn $c]
  if { $c1 != -1} {
      set d_node $c
      set d_num [lindex $gn $c1]
      } else {
          set d_node [lindex $dd $c2]
          set d_num $c
          }
 set c3 [expr $n-1]
 set d_last [lsearch $dd $d_node]

 if {$c3 == $d_last} {
# delete all lines of the last note (d_node)
 set d_line1 [concat $d_line1 $edgeFirst($d_node)]
 set d_line2 [concat $d_line2 $edgeSecond($d_node)]
 set d_line3 [concat $d_line1 $d_line2]

 foreach d_line $d_line1 {
   foreach d $dd {
     set l1 [lsearch $edgeSecond($d) $d_line]
     if { $l1 != -1 } {
      set edgeSecond($d) [lreplace $edgeSecond($d) $l1 $l1] 
     }
    }
  }

 foreach d_line $d_line2 {
   foreach d $dd {
     set l2 [lsearch $edgeFirst($d) $d_line]
     if { $l2 != -1 } { 
      set edgeFirst($d) [lreplace $edgeFirst($d) $l2 $l2] 
     }
    }
  }

 set edgeFirst($d_node) {}
 set edgeSecond($d_node) {}

#delete all the lines on graph
 foreach d_line $d_line3 {
   set l3 [lsearch $ll $d_line]
   set ll [lreplace $ll $l3 $l3]
   .relation.c delete $d_line
 }

# deleate the last note and number on graph
 set n [expr $n-1]
 set dd_num [expr $dd_num-1]
 .relation.c delete $d_num                   
 .relation.c delete $d_node                  

#adjust all the relations in dd dd1 gn
 set dd [lreplace $dd $c3 $c3]
 set dd1 [lreplace $dd1 $c3 $c3]
 set gn [lreplace $gn $c3 $c3]

#delete last X_coord. and Y_corod.in list node_X and list node_Y
 set drop_last $dd_num 
 set node_X [lreplace $node_X $drop_last $drop_last]
 set node_Y [lreplace $node_Y $drop_last $drop_last]
 }
}
focus .relation.c

#bind . <Control-c> {puts stdout $dd_num}
# puts stdout $node_X
# puts stdout $node_Y

