/*----------------------------------------------------------------------------*/
/* cre8btsp.c -- routine to create a file containing distances between a set  */
/*               of randomly-placed cities, for use with appbtsp.c, the blind */
/*               traveling salesman problem application accompanying sga and  */
/*               the Island Parallel Genetic Algorithm.                       */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "sga.h"

#define MAXNUMCITIES 32

int     numfields;
double  distbetween[MAXNUMCITIES][MAXNUMCITIES];
float   lengthofside;

main()
{
    void    make_distbetween_table();
    char    tablefile[13];
    int     i, j;
    static FILE *fpwrite = NULL;
    float randomseed;

    fprintf(stdout, "\n Welcome.  This program helps you create the table \n");
    fprintf(stdout, " distances between cities for the TSP problem.\n");
    fprintf(stdout, " Enter  random number seed, 0. < seed < 1.0\n");
    fflush(stdout);
    Randomize(&randomseed);
    fprintf(stdout, " Enter filename of intercity distbetween table: ");
    fscanf(stdin, "%13s", tablefile);
    if ((fpwrite = fopen(tablefile, "wb")) == NULL) {
	fprintf(stderr, " Error opening Distance Table File. \n");
	exit(-1);
    }
    fprintf(stdout, " Enter number of cities desired: ");
    fscanf(stdin, "%d", &numfields);
    if (numfields > MAXNUMCITIES) {
	fprintf(stderr, " Error.  Must revise MAXNUMCITIES and recompile to make that many cities.\n");
	exit(-1);
    }

    fprintf(stdout, " Enter length of one side of map containing all cities: ");
    fscanf(stdin, "%f", &lengthofside);

    make_distbetween_table();
    if (fwrite(&(numfields), sizeof(int), 1, fpwrite) != 1) {
	fprintf(stderr, "Error writing to City File.\n");
	exit(-1);
    }
    for (i = 0; i < numfields; i++)
	for (j = 0; j < numfields; j++) {	/* use FULL table for eval
						 * speed */
	    if (fwrite(&(distbetween[i][j]), sizeof(double), 1, fpwrite) != 1) {
		fprintf(stderr, " Error writing City File.\n");
		exit(-1);
	    }
	    fprintf(stdout, " distbetween[%d][%d] = %f\n", i, j
                    , distbetween[i][j]);
	}
    fclose(fpwrite);
}


void
make_distbetween_table()
{
    float   latitude[MAXNUMCITIES], longitude[MAXNUMCITIES];
    float   rndreal(), a, b;
    int     i, j;

    for (i = 0; i < numfields; i++) {
	latitude[i] = rndreal(0., lengthofside);
	longitude[i] = rndreal(0., lengthofside);
    }
    for (i = 0; i < numfields; i++) {
	for (j = 0; j < numfields; j++) {
	    a = longitude[i] - longitude[j];
	    b = latitude[i] - latitude[j];
	    distbetween[i][j] = sqrt(a * a + b * b);
	}
    }
}

void
advance_random()
/* Create next batch of 55 random numbers */
{
    int     j1;
    double  new_random;

    for (j1 = 0; j1 < 24; j1++) {
	new_random = oldrand[j1] - oldrand[j1 + 31];
	if (new_random < 0.0)
	    new_random = new_random + 1.0;
	oldrand[j1] = new_random;
    }
    for (j1 = 24; j1 < 55; j1++) {
	new_random = oldrand[j1] - oldrand[j1 - 24];
	if (new_random < 0.0)
	    new_random = new_random + 1.0;
	oldrand[j1] = new_random;
    }
}

void
Randomize(randomseed)
float   *randomseed;
/* Get seed number for random and start it up */
{
    int     j1;

    for (j1 = 0; j1 <= 54; j1++)
	oldrand[j1] = 0.0;
    jrand = 0;

    if (interactive) {
	do {
	    printf(" Enter random number seed, 0.0 to 1.0 -> ");
	    fscanf(stdin, "%f", randomseed);
	}
	while ((*randomseed < 0.0) || (*randomseed > 1.0));
    }
    else {
	fscanf(stdin, "%f", randomseed);
    }


    warmup_random(*randomseed);
}

float
randomperc()
/* Fetch a single random number between 0.0 and 1.0 - Subtractive Method */
/* See Knuth, D. (1969), v. 2 for details */
/* name changed from random() to avoid library conflicts on some machines*/
{
    jrand++;
    if (jrand >= 55) {
	jrand = 1;
	advance_random();
    }
    return ((float) oldrand[jrand]);
}

float
rndreal(lo, hi)
/* real random number between specified limits */
float   lo, hi;
{
    return ((randomperc() * (hi - lo)) + lo);
}

void
warmup_random(random_seed)
/* Get random off and running */
float   random_seed;
{
    int     j1, ii;
    double  new_random, prev_random;

    oldrand[54] = random_seed;
    new_random = 0.000000001;
    prev_random = random_seed;
    for (j1 = 1; j1 <= 54; j1++) {
	ii = (21 * j1) % 54;
	oldrand[ii] = new_random;
	new_random = prev_random - new_random;
	if (new_random < 0.0)
	    new_random = new_random + 1.0;
	prev_random = oldrand[ii];
    }

    advance_random();
    advance_random();
    advance_random();

    jrand = 0;
}
