/*----------------------------------------------------------------------------*/
/* cx.c --  Cycle crossover (for permutation-type problems), as described in  */
/*          Goldberg book.  Also includes FirstNth(), a function called by    */
/*          permutation-type mutation operators.  This assures that compatible*/
/*          permutation and crossover operators are loaded for a permutation- */
/*          type problem.  Returns jcross{0],[1] as 0.  */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <stdlib.h>

char which_crossover[] = "cx.c";

void
crossover(parent1, parent2, child1, child2, jcross)
unsigned *parent1, *parent2, *child1, *child2;
int     jcross[2];

/* Cross 2 parent strings, place in 2 child strings */
{
    int     i, j, k, cnt, realcnt;
    int     first;

    /* Do crossover with probability pcross */
    if (flip(pcross)) {
	ncross++;
	chromtointarray(intarray1, parent1);
	chromtointarray(intarray2, parent2);
	chromtointarray(intarraybak1, parent1);
	chromtointarray(intarraybak2, parent2);

        jcross[0] = 0;
        jcross[1] = 0;
	/* fprintf(stderr,"\n")    ; for     (i=0;i<numfields;i++)
	 * fprintf(stderr,"%3d ",intarray1[i])     ; fprintf(stderr,"\n")
	 * ; for     (i=0;i<numfields;i++) fprintf(stderr,"%3d
	 * ",intarray2[i]) ; */

	for (i = 0; i < numfields; i++)
	    tmparray[i] = 1;

	first = intarraybak1[0];
	k = intarraybak2[0];
	tmparray[0] = 0;
	while (1) {
	    for (j = 0; j < numfields; j++)
		if (k == intarraybak1[j]) {
		    k = intarraybak2[j];
		    tmparray[j] = 0;
		    break;
		}
	    if (first == k)
		break;
	}
	/* fprintf(stderr,"\n")    ; */
	for (i = 0; i < numfields; i++) {
	    /* fprintf(stderr," %d",tmparray[i])       ; */
	    if (tmparray[i]) {
		intarray1[i] = intarraybak2[i];
		intarray2[i] = intarraybak1[i];
	    }
	}

	/* fprintf(stderr,"\nResult:\n")   ; for     (i=0;i<numfields;i++)
	 * fprintf(stderr,"%3d ",intarray1[i])     ; fprintf(stderr,"\n")
	 * ; for     (i=0;i<numfields;i++) fprintf(stderr,"%3d
	 * ",intarray2[i]) ; fprintf(stderr,"\n")    ; bioskey(0)      ; */

	intarraytochrom(intarray1, child1);
	intarraytochrom(intarray2, child2);
    }
    else {
	for (k = 0; k < chromsize; k++) {
	    child1[k] = parent1[k];
	    child2[k] = parent2[k];
	}
	jcross[0] = 0;
	jcross[1] = 0;
    }
}

int
FirstNth(array, k, start, end)
int    *array;
int     k, start, end;
{
    int     cnt, i;

    cnt = 0;
    for (i = start; i <= end; i++)
	if (array[i])
	    cnt++;
    if (!cnt) {
	fprintf(stderr, "\nTemplate is wrong in FirstNth.\n");
	exit(-1);
    }
    cnt = k % cnt;
    for (i = start; i <= end; i++) {
	if (array[i]) {
	    if (!cnt)
		return i;
	    else
		cnt--;
	}
    }
    fprintf(stderr, "\nTemplate wrong in FirstNth.\n");
    exit(-1);
}
