/*----------------------------------------------------------------------------*/
/* mainmany.c -- main program for the GALOPP System                           */
/*----------------------------------------------------------------------------*/

/***********************************************/
/*                                             */
/* GALOPPS -- the "Genetic ALgorithm Optimized */
/*   for Portability and Parallelism" System   */
/*  (c) 1993,1994,1995,1996 Board of Trustees  */
/*  Michigan State University, East Lansing    */
/*        Written by Erik D. Goodman,          */
/*        Genetic Algorithms Research          */
/*           and Applications Group            */
/*  Michigan State University, East Lansing    */
/*                                             */
/*  Michigan State University, East Lansing    */
/*  THIS IS FREE SOFTWARE, AND IS DISTRIBUTED  */
/*  UNDER THE TERMS OF THE GNU PUBLIC SOFTWARE */
/*  LICENSE, LATEST VERSION.  A COPY OF THE    */
/*  LICENSE AGREEMENT IS IN FILE 'LICENSE'     */
/*  IN THE 'docs' SUBDIRECTORY.  THERE IS      */
/*  ABSOLUTELY NO WARRANTY, EXPRESSED OR       */
/*  IMPLIED, CONCERNING THIS SOFTWARE, AND ALL */
/*  RESPONSIBILITY AND RISK FOR ITS USE RESTS  */
/*  WITH THE USER.                             */
/*                                             */
/*         Release 3.2, April 20, 1996         */
/*                                             */
/*       Multiple Subpopulations Module        */
/*                                             */
/*              "Family Tree":                 */
/*                                             */
/*      Simple Genetic Algorithm - SGA         */
/*              Haploid Version                */
/* (c)    David Edward Goldberg  1986          */
/*            All Rights Reserved          --> */
/*        C translation by R.E. Smith      --> */
/*    v1.1 modifications by Jeff Earickson --> */
/*               Bug Fixes by                  */
/*       Erik D. Goodman and Wang Gang     --> */
/*            ESGA Releases 1.0, ...       --> */
/*      ESGA/IPGA System, Rel. 2.0, 2.01   --> */
/*           ESGA/IPGA Release 2.05        --> */
/*            GALOPPS Release 2.05         --> */
/*            GALOPPS Release 2.20         --> */
/*            GALOPPS Release 2.30         --> */
/*            GALOPPS Release 2.31         --> */
/*            GALOPPS Release 2.32         --> */
/*            GALOPPS Release 2.35         --> */
/*            GALOPPS Release 2.36         --> */
/*            GALOPPS Release 2.37         --> */
/*            GALOPPS Release 2.50         --> */
/*            GALOPPS Release 3.0          --> */
/*            GALOPPS Release 3.01         --> */
/*                                             */
/*   Current Release, GALOPPS, Release 3.2     */
/*                                             */
/***********************************************/

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#ifdef __MWERKS__
  #include <console.h>
#endif
#include "sga.h"

/* Henceforth, RESTART shall mean the reloading of a population from */
/* its previously saved checkpoint files, whether or not the run was */
/* started from "scratch" (user input -- interactive or file) OR not.*/

/* EXTEND shall mean the process of reinitiating a run to continue it, */
/* using saved checkpoint files, by providing the restartfileprefix to a */
/* newly-started copy of GALOPPS.  */

#ifdef PROTOTYPES_ACCEPTED
int commandline(int, char**);
#else
int commandline();
#endif
void input_overrides();
void global_overrides();
void plot();


main(argc, argv)
     int     argc;
     char   *argv[];
{
  char    c;
  BOOL    askrevise = TRUE;
 
  many = 1;

  /* Initialize timer */
  clock();

  /* process command line */
  #ifdef __MWERKS__
    argc = ccommand( &argv );
  #endif
  parameter_changes = commandline(argc, argv);

  /* print the author/copyright notice */
  copyright(1);

  /* Get inputs common to all subpopulations, set up checkpoint files */
  GetUniversalInputs();
  if (parameter_changes && !interactive)
    global_overrides();
  
  /* Call app_set_options() to let user set conditions for run. */
  app_set_options(); 

  for (cycle = 0; cycle < ncycles; cycle++) {
    if ((askrevise == TRUE) && extending) {
      c = 'n';
      if (interactive) {
	printf("Do you want to complete the run without");
	printf(" changing parameters? (n) -> ");
	fflush(stdout);
      }
      ffscanf("no_more_changes", infp, "%c", &c);
      
      if (c == 'y' || c == 'Y')
	askrevise = FALSE;
    }

    /* At start of each cycle, zero out the one_cycle_xxx totals, since */
    /* they are to accumulate quantities only during ONE cycle of all */
    /* pops running on THIS processor */
    local_cycle_sumfitness = 0.0;
    local_cycle_sum_best_fitness = 0.0;
    local_cycle_online_denominator = 0;
    local_cycle_best_fitness_count = 0;
    local_cycle_neval = 0;

    /* Start of main loop within a cycle */
    for (popno = startpopnum; popno <= finishpopnum; popno++) {

      subpopulation(&askrevise);

      globalstats();
      fflush(outfp);
      freeall();
      if (all_subpops_use_same_parameters)
	askrevise = FALSE;
    }

    askrevise = FALSE;
    /* End of cycle clean-up, reporting operations */
    EndOfCycle();
  }
}

void
subpopulation(askrevise)
     BOOL    *askrevise;
{
  struct individual *temp;
  FILE   *fopen();
  static int firstcall = 1;

  /* Set things up */
  initialize_parallel(askrevise);

  if (firstcall) {
    /* First call only, (startup or extending a run) proc. running subpop 0 */
    /* checks for existence of */
    /* .stt file containing the performance measures for all subpops. If */
    /* exists, use its values.  If it doesn't, create it and zero values.  */
    /* If things work correctly, only ONE process(or) (with subpop 0) will */
    /* create it.      */
    if (ReadCumStats() == FALSE) {
      fprintf(stderr, "\n***ERROR:  ReadCumStats failed.\n");
      exit(-1);
    }

    firstcall = 0;
    if (permproblem == 1)
      /* Malloc temporary space needed for permutation-type operators */
      MallocPermOperSpace();
  }

  startgen = gen;
  maxgen = gen + genspercycle;  /* gen is last one run, from file */
  for (gen = startgen; gen < maxgen; gen++) {
    if (quiet == 0) 
      fprintf(outfp, "\nSubpopulation %d, GENERATION %d->%d\n",
	      popno, gen, maxgen);

    /* application dependent routines */
    application();

    /* create a new generation */
    generation();

    /* compute fitness statistics on new populations */
    statistics(newpop);
    windowend++;                /* advance window for window scaling */

    /* report results for new generation */
    plot();
    if (quiet < 2)
      report();
    else
      app_quiet_report();

    /* advance the generation (move new guys into oldpop to keep) */
    temp = oldpop;
    oldpop = newpop;
    newpop = temp;

    /* Allow user to make changes if population performance or */
    /* convergence measures meet certain user-specified criteria. */
    ModifyRunBasedOnConvergence();
  } /* END OF CYCLE for one subpop */

  if (quiet < 2)
    /* Note:  print below occurs before all_pops_neval is updated for */
    /* the evals done this cycle, so add them to the total to print. */
    fprintf(outfp,"**Subpop %d, Gen %d done; Total eval.'s all pops: %ld\n", 
	    popno, gen, all_pops_neval + local_cycle_neval);

  write_checkpoint();
}


/* Called from main().  Performs output and bookkeeping operations at end */
/* of each cycle */
void
EndOfCycle()
{
  /* Print out global performance measures, unless very quiet mode */
  if (quiet < 2) {
    fprintf(outfp, "\n Performance Across All Populations Recorded To Date:");
    fprintf(outfp, "\n On-Line= %f, Off-line= %f\n"
	    , all_pops_sumfitness / all_pops_online_denominator
	    , all_pops_sum_best_fitness / all_pops_best_fitness_count);
    fprintf(outfp, "\n *** End of Cycle ***\n\n");
  }

  if (cycle == 0) {
    if (!strlen(checkptfileprefix))
      strcpy(restartfileprefix, "sgackp");
    else
      strcpy(restartfileprefix, checkptfileprefix);
  }

  for (popno = startpopnum; popno <= finishpopnum; popno++)  {
    CombineFileName(restarthdrfilename, restartfileprefix, popno, ".ckp");
    update_rstate(restarthdrfilename, 1);
    CombineFileName(restartindfilename, restartfileprefix, popno, ".ind");
    update_rstate(restartindfilename, 1);
  }

  /* Population runs completed for cycle... update cumulative stats */
  /* for ALL SUBPOPULATIONS, ALL PROCESSORS.  This involves locking */
  /* the .stt file, then reading from it, incrementing the */
  /* all_pops values read in using local_cycle values, then writing */
  /* out the updated all_pops values, then unlocking the file.  */
  if (UpdateCumStats() == FALSE) {
    fprintf(stderr, "***OOPS, UpdateCumStats Failed in Mainmany.c");
    exit(-1);
  }
  fflush(outfp);
}


/* Allow the user the opportunity to make changes in parameters, create
   new individuals, terminate run, or take other actions, based on any of
   the performance or convergence measures for the population. */
void
ModifyRunBasedOnConvergence()
{
  static int converg_cond_flag; /* set in app_decide_if_converged, */
  /* 0 = keep going; */
  /* 1 = reinitialize number_to_replace guys to */
  /*     random values; */
  /* 2 = call app_when_converged() to let user  */
  /*     do whatever wants. */
  static int number_to_replace = 0; /* may be set by user in routine */
  /* app_decide_if_converged */
  static int starting_guy_index; /* index in oldpop where start random */
  /* initialization or reinitialization. */
  int i, newbest;

  /* Allow user chance to decide converged, replace some or all before */
  /* advancing generation, or quit, or whatever wants. */
  converg_cond_flag = 0;
  app_decide_if_converged(&converg_cond_flag, &number_to_replace);
  switch (converg_cond_flag) {
  case 0:        /* Just keep going */
    break;
  case 1:        /* keep "starting_guy_index" guys, randomly reinitialize */
		 /* all the rest. */
    starting_guy_index = popsize - number_to_replace;
    if (quiet < 3) {
      fprintf(outfp, "\n *** Subpopulation %d converged ***", popno);
      fprintf(outfp, "\n Reinitializing %d individuals.\n"
	      , number_to_replace);
    }
    if (elitism && starting_guy_index > 0) {
      /* Be sure last copy of best isn't overwritten -- store him in */
      /* slot 0, reinitialize later guys */
      oldpop[0] = oldpop[bestnow];
      bestnow = 0;
    }

    if (user_supplied_initialization)
      app_user_init_pop(starting_guy_index);
    else if (permproblem)
      initpermpop(starting_guy_index);
    else if (alpha_size != 2)
      init_non_binary(starting_guy_index);
    else if (superuniform_initialization)
      init_superuniform_pop(starting_guy_index);
    else
      init_random_pop(starting_guy_index);

    /* Record NEW best guy, whether elitism on or not. */
    newbest = bestnow;
    for (i=0;i<popsize;i++) {
      if (oldpop[i].init_fitness > oldpop[bestnow].init_fitness)
	newbest = i;
    }
    if (newbest != bestnow) {
      bestnow = newbest;
      bestfit.init_fitness = oldpop[bestnow].init_fitness;
      bestfit.neval = oldpop[bestnow].neval;
      bestfit.generation = gen;
      for (i=0;i<chromsize;i++)
	bestfit.chrom[i] = oldpop[bestnow].chrom[i];
      sys_util_copy_utility(bestfit.utility, oldpop[bestnow].utility);
    }
    break;

  case 2:
    app_when_converged();
    break;
  default:
    fprintf(stderr, "\n *** Error in app convergence logic or subpop() ***.\n");
    exit(-111);
  }
}
