/*----------------------------------------------------------------------------*/
/* pmx.c --  PMX (Partially Matched Crossover) operator, for permutation-type */
/*                 problems, as described in Goldberg book.  File also        */
/*                 includes FirstNth(), a routine called by permutation-type  */
/*                 mutation operators, to assure that user loads compatible   */
/*                 crossover and mutation operators when doing permutation-   */
/*                 type problems */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <stdlib.h>

char which_crossover[] = "pmx.c";

void
crossover(parent1, parent2, child1, child2, jcross)
/* Perform PMX, "partially matched crossover," as described in Goldberg book */
unsigned *parent1, *parent2, *child1, *child2;
int     jcross[2];

/* Cross 2 parent strings, place in 2 child strings */
{
    int     i, j, k, cnt, realcnt;
    int     start, end;

    /* Do crossover with probability pcross */
    if (flip(pcross)) {
	ncross++;
	chromtointarray(intarray1, parent1);
	chromtointarray(intarraybak1, parent1);
	chromtointarray(intarray2, parent2);
	chromtointarray(intarraybak2, parent2);

	j = rnd(0, numfields - 1);
	k = rnd(0, numfields - 1);
	start = (j < k) ? j : k;
	end = (j > k) ? j : k;

	jcross[0] = start;
	jcross[1] = end;

	/* fprintf(stderr,"\n")    ; for     (i=0;i<numfields;i++)
	 * fprintf(stderr,"%d ",intarray1[i])      ; fprintf(stderr,"\n")
	 * ; for     (i=0;i<numfields;i++) fprintf(stderr,"%d ",intarray2[i])
	 * ; fprintf(stderr,"\nstart:%d end:%d",start,end)   ; */
	/* now generate child1. */
	for (i = start; i <= end; i++)
	    tmparray[i] = 1;
	for (i = 0; i < start; i++)
	    for (j = start; j <= end; j++)
		if (intarraybak1[i] == intarraybak2[j]) {
		    intarray1[i] = intarraybak1[j];
		    tmparray[j] = 0;
		}
	for (i = end + 1; i < numfields; i++)
	    for (j = start; j <= end; j++)
		if (intarraybak1[i] == intarraybak2[j]) {
		    intarray1[i] = intarraybak1[j];
		    tmparray[j] = 0;
		}
	for (i = start; i <= end; i++)
	    if (tmparray[i] == 0)
		intarray1[i] = intarraybak2[i];

	/* now generate child2. */
	for (i = start; i <= end; i++)
	    tmparray[i] = 1;
	for (i = 0; i < start; i++)
	    for (j = start; j <= end; j++)
		if (intarraybak2[i] == intarraybak1[j]) {
		    intarray2[i] = intarraybak2[j];
		    tmparray[j] = 0;
		}
	for (i = end + 1; i < numfields; i++)
	    for (j = start; j <= end; j++)
		if (intarraybak2[i] == intarraybak1[j]) {
		    intarray2[i] = intarraybak2[j];
		    tmparray[j] = 0;
		}
	for (i = start; i <= end; i++)
	    if (tmparray[i] == 0)
		intarray2[i] = intarraybak1[i];
	/* fprintf(stderr,"\nResult:\n")   ; for     (i=0;i<numfields;i++)
	 * fprintf(stderr,"%d ",intarray1[i])      ; fprintf(stderr,"\n")
	 * ; for     (i=0;i<numfields;i++) fprintf(stderr,"%d ",intarray2[i])
	 * ; fprintf(stderr,"\n")    ; bioskey(0)      ; */
	intarraytochrom(intarray1, child1);
	intarraytochrom(intarray2, child2);
    }
    else {
	for (k = 0; k < chromsize; k++) {
	    child1[k] = parent1[k];
	    child2[k] = parent2[k];
	}
	jcross[0] = 0;
	jcross[1] = 0;
    }
}

int
FirstNth(array, k, start, end)
int    *array;
int     k, start, end;
{
    int     cnt, i;

    cnt = 0;
    for (i = start; i <= end; i++)
	if (array[i])
	    cnt++;
    if (!cnt) {
	fprintf(stderr, "\nTemplate is wrong in FirstNth.\n");
	exit(-1);
    }
    cnt = k % cnt;
    for (i = start; i <= end; i++) {
	if (array[i]) {
	    if (!cnt)
		return i;
	    else
		cnt--;
	}
    }
    fprintf(stderr, "\nTemplate wrong in FirstNth.\n");
    exit(-1);
}
