/*----------------------------------------------------------------------------*/
/* suselect.c - contains select_memory, select_free,                          */
/*         preselect, and select, for stochastic universal sampling method    */
/*         Reference is James Baker, Proceedings, Second International Conf.  */
/*         on Genetic Algorithms, pp.14-21, 1987, "Reducing Bias and          */
/*         Inefficency in the Selection Algorithm.  Coded for SGA by Erik     */
/*         Goodman, October 30, 1993.                                         */
/*----------------------------------------------------------------------------*/

#include "external.h"

extern float randomperc();
static int *choices, nremain;

char which_selection[] = "suselect.c";

void
select_memory()
{
  /* allocates aux. memory for stochastic universal sampling selection */

  int nbytes;

  nbytes = popsize * sizeof(int);
  if ((choices = (int *) malloc(nbytes)) == NULL)
    nomemory("choices");
}

void
select_free()
{
  /* frees aux. memory for stochastic universal sampling selection */
  free(choices);
  choices = NULL;
}

void
preselect()
/* preselection for stochastic universal sampling method */
{
  int     j, k;
  float   pointer, sum;

  if (fit_avg == 0.) {
    for (j = 0; j < popsize; j++)
      choices[j] = j;
  }
  else {
    k = 0;
    pointer = randomperc();
    sum = 0.0;
    for (j = 0; j < popsize; j++) {
      for (sum += (oldpop[j].fitness / fit_avg); sum > pointer; pointer++) {
	choices[k++] = j;
	if(k >= popsize) 
          break;
      }
      if(k >= popsize)
        break;
    }
  }
  nremain = popsize - 1;
}


int
selection()
/* selection using stochastic universal sampling method */
{
  int     jpick, slect;

  jpick = rnd(0, nremain);
  slect = choices[jpick];
  choices[jpick] = choices[nremain];
  nremain--;
  return (slect);
}
