/*----------------------------------------------------------------------------*/
/* tselect.c - Tournament selection                                           */
/* Contributed by Hillol Kargupta, Dept of Engr. Mechanics, U. of Alabama     */
/* Input logic updated, moved to initialization, April 3, 1995, Erik Goodman. */
/*----------------------------------------------------------------------------*/

#include "external.h"

static void reset();
char which_selection[] = "tselect.c";

static int *tourneylist, tourneypos;
static int firsttime = 1;

/* tournament list, position in list */

void
select_memory()
{
    unsigned nbytes;

    nbytes = popsize * sizeof(int);
    if ((tourneylist = (int *) malloc(nbytes)) == NULL)
	nomemory("tourneylist");
}


void
select_free()
{
    free(tourneylist);
    tourneylist = NULL;
}



void
preselect()
{
    reset();
    tourneypos = 0;
}


int
selection()
{
    int     pick, winner, i;

    /* If remaining members not enough for a tournament, then reset list */
    if ((popsize - tourneypos) < tourneysize) {
	reset();
	tourneypos = 0;
    }

    /* Select tourneysize structures at random and conduct a tournament */
    winner = tourneylist[tourneypos];
    for (i = 1; i < tourneysize; i++) {
	pick = tourneylist[i + tourneypos];
	if (oldpop[pick].fitness > oldpop[winner].fitness)
	    winner = pick;
    }

    /* Update tourneypos */
    tourneypos += tourneysize;
    return (winner);
}


void
reset()
/* Shuffles the tourneylist at random */
{
    int     i, rand1, rand2, temp;

    for (i = 0; i < popsize; i++)
	tourneylist[i] = i;

    for (i = 0; i < popsize; i++) {
	rand1 = rnd(i, popsize - 1);
	rand2 = rnd(i, popsize - 1);
	temp = tourneylist[rand1];
	tourneylist[rand1] = tourneylist[rand2];
	tourneylist[rand2] = temp;
    }
}
