/*-------------------------------------------------------------------------*/
/* uniformx.c -- Uniform crossover operator for ordinary (binary string or */
/* larger alphabet (non-permutation) problems.  Not for permutation-type   */
/* representations.  Uses simpler, faster logic for binary chromosomes.    */
/* Modified 4/96 by EDG to handle different length fields, crossing only   */
/* at field boundaries.                                                    */
/*-------------------------------------------------------------------------*/

#include "external.h"

char which_crossover[] = "unifx.c";

void crossover(parent1, parent2, child1, child2, jcross)
unsigned *parent1, *parent2, *child1, *child2;
int jcross[2];

{

 int counter1, counter2, bitsleft, UintSize, fieldnum, lclfldlength; 
 int prevfldlength;
 int rememberflip;
 unsigned mask;

 jcross[0]=0; /* not used in uniform crossover, so report 0 */
 jcross[1]=0; /* not used in uniform crossover, so report 0 */

 if (flip(pcross)) {
    ncross++;
    if (alpha_size == 2) {
        /* BINARY ALPHABET, CROSS ANYWHERE */
	for (counter1=0; counter1<chromsize; counter1++) {
	    child1[counter1]=0;
	    child2[counter1]=0;
	    mask=1;
	    for (counter2=1;
		 (counter2<=UINTSIZE)
		    &&(counter2+(counter1)*UINTSIZE<=(unsigned)lchrom);
		 counter2++) {

		if (flip(0.5)) {
		    child1[counter1]=child1[counter1] | (mask & parent1[counter1]);
		    child2[counter1]=child2[counter1] | (mask & parent2[counter1]);
		}
		else {
		    child1[counter1]=child1[counter1] | (mask & parent2[counter1]);
		    child2[counter1]=child2[counter1] | (mask & parent1[counter1]);
		}
		mask = mask << 1;
	    }
	}
    }
    else {
	/* non-binary fields, alphabet size doesn't necessarily use all   */
	/* possible bit combinations, must cross on field boundaries,     */
	/* to perform valid crossovers.  Requires care at unsigned word   */
	/* boundaries.                                                    */
	bitsleft = 0;
	child1[0] = 0;
	child2[0] = 0;
	fieldnum = 0;
	lclfldlength = field_ends[0];
        prevfldlength = 0;
	for (counter1=0; counter1<chromsize; counter1++) {   /* which word */
	    for (counter2=bitsleft;                /* starting bit in word */
		 (counter2<UINTSIZE)
		    &&(counter2+(counter1)*UINTSIZE<lchrom);
                 counter2 += prevfldlength) {
	        mask = 1;
                mask = (mask << lclfldlength) - 1;
	        mask = mask << counter2;

		if ((rememberflip = flip(0.5))) {
		    child1[counter1]=child1[counter1] | (mask & parent1[counter1]);
		    child2[counter1]=child2[counter1] | (mask & parent2[counter1]);
		}
		else {
		    child1[counter1]=child1[counter1] | (mask & parent2[counter1]);
		    child2[counter1]=child2[counter1] | (mask & parent1[counter1]);
		}
                fieldnum++;
                prevfldlength = lclfldlength;
		lclfldlength = field_ends[fieldnum] - field_ends[fieldnum - 1];
	    }
	    /* field may cross word boundary -- fix up if does (else NOP) */
	    bitsleft = counter2 % UINTSIZE;
            mask = 1;
	    mask = (mask << bitsleft) - 1;
	    /* Below, finishing current field in beginning of next word */
	    if (counter1 < chromsize - 1) {
		child1[counter1+1] = 0;
		child2[counter1+1] = 0;

		if (rememberflip) {
		    child1[counter1+1]=child1[counter1+1] | (mask & parent1[counter1+1]);
		    child2[counter1+1]=child2[counter1+1] | (mask & parent2[counter1+1]);
		}
		else {
		    child1[counter1+1]=child1[counter1+1] | (mask & parent2[counter1+1]);
		    child2[counter1+1]=child2[counter1+1] | (mask & parent1[counter1+1]);
		}
	    }
            /* Now working with all but "bitsleft" bits of new word. */
            
	}
    }
 }
 else {
    for (counter1 = 0; counter1 < chromsize; counter1++) {
	child1[counter1] = parent1[counter1];
	child2[counter1] = parent2[counter1];
    }
 }

}


